from __future__ import annotations


__lazy_modules__ = {'packaging', 'packaging.requirements', f'{__spec__.parent}._compat'}

import sys

import packaging.requirements

from ._compat import importlib


TYPE_CHECKING = False

if TYPE_CHECKING:
    from collections.abc import Iterator
    from collections.abc import Set as AbstractSet


def check_dependency(
    req_string: str, ancestral_req_strings: tuple[str, ...] = (), parent_extras: AbstractSet[str] = frozenset()
) -> Iterator[tuple[str, ...]]:
    """
    Verify that a dependency and all of its dependencies are met.

    :param req_string: Requirement string
    :param parent_extras: Extras (eg. "test" in myproject[test])
    :yields: Unmet dependencies
    """
    yield from _check_dependency(req_string, ancestral_req_strings, parent_extras, set())


def _check_dependency(
    req_string: str, ancestral_req_strings: tuple[str, ...], parent_extras: AbstractSet[str], seen: set[str]
) -> Iterator[tuple[str, ...]]:
    req = packaging.requirements.Requirement(req_string)
    normalised_req_string = str(req)

    # ``seen`` holds requirements already verified as fully satisfied (nothing
    # yielded), so shared subtrees of a diamond dependency graph are walked once.
    if normalised_req_string in seen:
        return

    # ``Requirement`` doesn't implement ``__eq__`` so we cannot compare reqs for
    # equality directly but the string representation is stable.
    if normalised_req_string in ancestral_req_strings:
        # cyclical dependency, already checked.
        return

    if req.marker:
        extras = frozenset(('',)).union(parent_extras)
        # a requirement can have multiple extras but ``evaluate`` can
        # only check one at a time.
        if all(not req.marker.evaluate(environment={'extra': e}) for e in extras):
            # if the marker conditions are not met, we pretend that the
            # dependency is satisfied.
            return

    try:
        dist = importlib.metadata.distribution(req.name)
    except importlib.metadata.PackageNotFoundError:
        # dependency is not installed in the environment.
        yield (*ancestral_req_strings, normalised_req_string)
        return

    if req.specifier and not req.specifier.contains(dist.version, prereleases=True):
        # the installed version is incompatible.
        yield (*ancestral_req_strings, normalised_req_string)
        return

    satisfied = True
    for other_req_string in dist.requires or ():
        # yields transitive dependencies that are not satisfied.
        for unmet in _check_dependency(other_req_string, (*ancestral_req_strings, normalised_req_string), req.extras, seen):
            satisfied = False
            yield unmet

    # unmet requirements are not memoised: they must be reported once per
    # dependency chain that reaches them.
    if satisfied:
        seen.add(normalised_req_string)


def format_unmet_dependencies(unmet: AbstractSet[tuple[str, ...]]) -> str:
    body = ''.join(_format_missing_dependency(chain) for chain in sorted(unmet))
    return f'Unmet dependencies (checked against {sys.executable}):{body}'


def _format_missing_dependency(dep_chain: tuple[str, ...]) -> str:
    requirement = packaging.requirements.Requirement(dep_chain[-1])
    try:
        found = importlib.metadata.distribution(requirement.name).version
    except importlib.metadata.PackageNotFoundError:
        found = 'not installed'
    wanted = str(requirement.specifier) if requirement.specifier else 'any'
    return f'\n\t{_format_dep_chain(dep_chain)}\n\t\twanted: {wanted}\n\t\tfound: {found}'


def _format_dep_chain(dep_chain: tuple[str, ...]) -> str:
    return ' -> '.join(dep.partition(';')[0].strip() for dep in dep_chain)



from __future__ import annotations

import collections.abc
import datetime
import os
import typing


__all__ = ['ConfigSettings', 'Distribution', 'JSONValue', 'StrPath', 'SubprocessRunner', 'TOMLValue']

ConfigSettings = collections.abc.Mapping[str, str | collections.abc.Sequence[str]]
Distribution = typing.Literal['sdist', 'wheel', 'editable']

StrPath = str | os.PathLike[str]

# A decoded JSON value, as produced by ``json.load``. Uses the covariant ``Sequence``/``Mapping``
# so concrete literals (e.g. ``str | list[str]``) are also assignable to it.
JSONValue = str | int | float | bool | None | collections.abc.Sequence['JSONValue'] | collections.abc.Mapping[str, 'JSONValue']

# A value as produced by ``tomllib`` when parsing ``pyproject.toml``. Uses the covariant
# ``Sequence``/``Mapping`` so concrete literals (e.g. ``dict[str, list[str]]``) are assignable.
TOMLValue = (
    str
    | int
    | float
    | bool
    | datetime.datetime
    | datetime.date
    | datetime.time
    | collections.abc.Sequence['TOMLValue']
    | collections.abc.Mapping[str, 'TOMLValue']
)

TYPE_CHECKING = False

if TYPE_CHECKING:
    from pyproject_hooks import SubprocessRunner
else:
    SubprocessRunner = collections.abc.Callable[
        [collections.abc.Sequence[str], str | None, collections.abc.Mapping[str, str] | None], None
    ]



# SPDX-License-Identifier: MIT

"""The projects that the integration tests build, and the cache of their archives.

``tests/test_integration.py`` imports this module, and CI runs it as a script to fill the cache once for the whole test
matrix. It uses the standard library only, so CI can run it before an install.

"""

from __future__ import annotations

import http.client
import os
import shutil
import time
import urllib.request

from pathlib import Path


INTEGRATION_SOURCES: dict[str, tuple[str, str]] = {
    'dateutil': ('dateutil/dateutil', '2.9.0'),
    'pip': ('pypa/pip', '25.0.1'),
    'Solaar': ('pwr-Solaar/Solaar', '1.1.14'),
    'flit': ('pypa/flit', '3.12.0'),
}

STORE_DIR = Path(__file__).resolve().parents[1] / '.integration-sources'
ATTEMPTS = 3
NETWORK_ERRORS = (OSError, http.client.HTTPException)


def download_archive(name: str) -> tuple[Path, str]:
    """Put the archive in the cache and return its path and version. Hold a lock around this to make it parallel safe."""
    github_org_repo, version = INTEGRATION_SOURCES[name]
    target = STORE_DIR / f'{name}-{version}.tar.gz'
    if target.exists():
        return target, version

    url = f'https://github.com/{github_org_repo}/archive/{version}.tar.gz'
    STORE_DIR.mkdir(exist_ok=True)
    # write to a temporary name, because GitHub can drop a large archive request and leave a truncated file behind
    partial = target.with_suffix('.part')
    for attempt in range(1, ATTEMPTS + 1):
        try:
            with urllib.request.urlopen(url) as request, partial.open('wb') as file_handler:
                shutil.copyfileobj(request, file_handler)
        except NETWORK_ERRORS as exception:  # noqa: PERF203
            partial.unlink(missing_ok=True)
            if attempt == ATTEMPTS:
                raise
            print(f'attempt {attempt}/{ATTEMPTS} failed for {url}: {exception}')
            time.sleep(2**attempt)
        else:
            break
    os.replace(partial, target)
    return target, version


def main() -> None:
    failed = []
    for name in INTEGRATION_SOURCES:
        try:
            print(f'ready: {download_archive(name)[0].name}')
        except NETWORK_ERRORS as exception:  # noqa: PERF203
            print(f'failed: {name}: {exception}')
            failed.append(name)
    if failed:
        print(f'::warning::could not pre-download {", ".join(failed)}; the tests download these themselves')


if __name__ == '__main__':
    main()



# SPDX-License-Identifier: MIT

from __future__ import annotations

import contextlib
import hashlib
import io
import json
import os
import pathlib
import re
import subprocess
import sys
import tarfile
import unittest.mock
import venv
import zipfile

from collections.abc import Callable, Generator
from typing import TYPE_CHECKING, Protocol, TypedDict

import pytest
import pytest_mock

import build
import build.__main__
import build._ctx
import build.env

from build._compat import importlib as _importlib


if TYPE_CHECKING:
    from conftest import SubTests

    from build._types import SubprocessRunner


pytestmark = pytest.mark.contextvars


build_open_owner = 'builtins'

cwd = os.getcwd()
out = os.path.join(cwd, 'dist')

ANSI_STRIP = re.compile(r'\x1B(?:[@-Z\\-_]|\[[0-?]*[ -/]*[@-~])')

JSONValue = str | int | list['JSONValue'] | dict[str, 'JSONValue']


class BuildKwargs(TypedDict):
    distributions: list[str]
    config_settings: dict[str, JSONValue]
    isolation: bool
    skip_dependency_check: bool
    dependency_constraints_txt: str | None
    installer: str
    env_dir: str | None


def make_kwargs(
    *,
    distributions: list[str] | None = None,
    config_settings: dict[str, JSONValue] | None = None,
    isolation: bool = True,
    skip_dependency_check: bool = False,
    dependency_constraints_txt: str | None = None,
    installer: str = 'pip',
    env_dir: str | None = None,
) -> BuildKwargs:
    return {
        'distributions': distributions if distributions is not None else ['wheel'],
        'config_settings': config_settings if config_settings is not None else {},
        'isolation': isolation,
        'skip_dependency_check': skip_dependency_check,
        'dependency_constraints_txt': dependency_constraints_txt,
        'installer': installer,
        'env_dir': env_dir,
    }


@pytest.mark.parametrize(
    ('cli_args', 'build_args', 'build_kwargs', 'hook'),
    [
        pytest.param([], (cwd, out), make_kwargs(), 'build_package_via_sdist', id='defaults'),
        pytest.param(['-n'], (cwd, out), make_kwargs(isolation=False), 'build_package_via_sdist', id='no-isolation'),
        pytest.param(['-s'], (cwd, out), make_kwargs(distributions=['sdist']), 'build_package', id='sdist'),
        pytest.param(['-w'], (cwd, out), make_kwargs(), 'build_package', id='wheel'),
        pytest.param(
            ['-s', '-w'], (cwd, out), make_kwargs(distributions=['sdist', 'wheel']), 'build_package', id='sdist-and-wheel'
        ),
        pytest.param(
            ['source'], ('source', os.path.join('source', 'dist')), make_kwargs(), 'build_package_via_sdist', id='srcdir'
        ),
        pytest.param(['-o', 'out'], (cwd, 'out'), make_kwargs(), 'build_package_via_sdist', id='outdir'),
        pytest.param(['source', '-o', 'out'], ('source', 'out'), make_kwargs(), 'build_package_via_sdist', id='srcdir-outdir'),
        pytest.param(
            ['-x'], (cwd, out), make_kwargs(skip_dependency_check=True), 'build_package_via_sdist', id='skip-dependency-check'
        ),
        pytest.param(
            ['--installer', 'uv'], (cwd, out), make_kwargs(installer='uv'), 'build_package_via_sdist', id='installer'
        ),
        pytest.param(
            ['-C--flag1=', '-C--flag2='],
            (cwd, out),
            make_kwargs(config_settings={'--flag1': '', '--flag2': ''}),
            'build_package_via_sdist',
            id='config-empty-values',
        ),
        pytest.param(
            ['-C--flag=value'],
            (cwd, out),
            make_kwargs(config_settings={'--flag': 'value'}),
            'build_package_via_sdist',
            id='config-single',
        ),
        pytest.param(
            ['-C--flag1=value', '-C--flag2=other_value', '-C--flag2=extra_value'],
            (cwd, out),
            make_kwargs(config_settings={'--flag1': 'value', '--flag2': ['other_value', 'extra_value']}),
            'build_package_via_sdist',
            id='config-repeated',
        ),
        pytest.param(
            ['--config-json={"one": 1, "two": [2, 3], "three": {"in": "out"}}'],
            (cwd, out),
            make_kwargs(config_settings={'one': 1, 'two': [2, 3], 'three': {'in': 'out'}}),
            'build_package_via_sdist',
            id='config-json',
        ),
        pytest.param(
            ['--config-json', '{"outer": {"inner": {"deeper": 2}}}'],
            (cwd, out),
            make_kwargs(config_settings={'outer': {'inner': {'deeper': 2}}}),
            'build_package_via_sdist',
            id='config-json-nested',
        ),
        pytest.param(['--config-json', '{}'], (cwd, out), make_kwargs(), 'build_package_via_sdist', id='config-json-empty'),
        pytest.param(
            ['--dependency-constraints-txt', 'contraints.txt'],
            (cwd, out),
            make_kwargs(dependency_constraints_txt='contraints.txt'),
            'build_package_via_sdist',
            id='dependency-constraints-txt',
        ),
        pytest.param(
            ['--env-dir', 'build-env'],
            (cwd, out),
            make_kwargs(env_dir='build-env'),
            'build_package_via_sdist',
            id='env-dir',
        ),
    ],
)
def test_parse_args(
    mocker: pytest_mock.MockerFixture,
    cli_args: list[str],
    build_args: tuple[str, str],
    build_kwargs: BuildKwargs,
    hook: str,
) -> None:
    build_package = mocker.patch('build.__main__.build_package', autospec=True, return_value=['something'])
    build_package_via_sdist = mocker.patch('build.__main__.build_package_via_sdist', autospec=True, return_value=['something'])

    build.__main__.main(cli_args)

    if hook == 'build_package':
        build_package.assert_called_with(*build_args, **build_kwargs)
    elif hook == 'build_package_via_sdist':
        build_package_via_sdist.assert_called_with(*build_args, **build_kwargs, sdist_extract_dir=None)
    else:  # pragma: no cover
        msg = f'Unknown hook {hook}'
        raise ValueError(msg)


def test_env_dir_flag_forwarded(mocker: pytest_mock.MockerFixture) -> None:
    build_package_via_sdist = mocker.patch('build.__main__.build_package_via_sdist', return_value=['something'])

    build.__main__.main(['--env-dir', 'build-env'])

    assert build_package_via_sdist.call_args.kwargs['env_dir'] == 'build-env'


def test_env_dir_conflicts_with_no_isolation(capsys: pytest.CaptureFixture[str]) -> None:
    with pytest.raises(SystemExit):
        build.__main__.main(['--env-dir', 'build-env', '--no-isolation'])

    assert '--env-dir: not allowed with --no-isolation' in capsys.readouterr().err


def test_prog() -> None:
    out = io.StringIO()

    with pytest.raises(SystemExit), contextlib.redirect_stdout(out):
        build.__main__.main(['--help'], prog='something')

    assert out.getvalue().startswith('usage: something [-h]')


def test_version(capsys: pytest.CaptureFixture[str]) -> None:
    with pytest.raises(SystemExit):
        build.__main__.main(['--version'])
    out, _ = capsys.readouterr()
    assert out.startswith(f'build {build.__version__}')


@pytest.mark.isolated
def test_build_isolated(mocker: pytest_mock.MockerFixture, package_test_flit: str) -> None:
    build_cmd = mocker.patch('build.ProjectBuilder.build', return_value='something')
    required_cmd = mocker.patch(
        'build.ProjectBuilder.get_requires_for_build',
        side_effect=[
            {'dep1', 'dep2'},
        ],
    )
    mocker.patch('build.__main__._error')
    install = mocker.patch('build.env.DefaultIsolatedEnv.install')

    build.__main__.build_package(package_test_flit, '.', ['sdist'])

    install.assert_any_call({'flit_core >=2,<4'}, _fresh=True)

    required_cmd.assert_called_with('sdist', None)
    install.assert_any_call({'dep1', 'dep2'})

    build_cmd.assert_called_with('sdist', '.', None)


def test_build_no_isolation_check_deps_empty(mocker: pytest_mock.MockerFixture, package_test_flit: str) -> None:
    # check_dependencies = []
    build_cmd = mocker.patch('build.ProjectBuilder.build', return_value='something')
    mocker.patch('build.ProjectBuilder.check_dependencies', return_value=[])

    build.__main__.build_package(package_test_flit, '.', ['sdist'], isolation=False)

    build_cmd.assert_called_with('sdist', '.', None)


def test_build_package_passes_config_settings_to_build(mocker: pytest_mock.MockerFixture, package_test_flit: str) -> None:
    build_cmd = mocker.patch(
        'build.__main__._build',
        autospec=True,
        side_effect=[
            os.path.join('dist', 'test_flit-1.0.0.tar.gz'),
            os.path.join('dist', 'test_flit-1.0.0-py3-none-any.whl'),
        ],
    )
    config_settings = {'--flag': 'value'}

    built = build.__main__.build_package(
        package_test_flit,
        '.',
        ['sdist', 'wheel'],
        config_settings=config_settings,
        isolation=False,
        skip_dependency_check=True,
        dependency_constraints_txt=pathlib.Path('constraints.txt'),
        installer='uv',
    )

    assert built == ['test_flit-1.0.0.tar.gz', 'test_flit-1.0.0-py3-none-any.whl']
    build_cmd.assert_has_calls(
        [
            unittest.mock.call(
                False, package_test_flit, '.', 'sdist', config_settings, True, pathlib.Path('constraints.txt'), 'uv', None
            ),
            unittest.mock.call(
                False, package_test_flit, '.', 'wheel', config_settings, True, pathlib.Path('constraints.txt'), 'uv', None
            ),
        ]
    )


def test_build_package_via_sdist_passes_config_settings_to_build(mocker: pytest_mock.MockerFixture) -> None:
    build_cmd = mocker.patch(
        'build.__main__._build',
        autospec=True,
        side_effect=[
            os.path.join('dist', 'demo-1.0.0.tar.gz'),
            os.path.join('dist', 'demo-1.0.0-py3-none-any.whl'),
        ],
    )
    mocker.patch('build.__main__.tempfile.mkdtemp', return_value='temp-sdist-dir')
    rmtree = mocker.patch('build.__main__.shutil.rmtree')
    tar_open = mocker.patch('build.__main__.tar_open')
    mocker.patch('build.__main__._validate_sdist_archive', return_value='demo-1.0.0')
    mocker.patch('build.__main__._ctx.log')
    config_settings = {'--flag': 'value'}

    built = build.__main__.build_package_via_sdist(
        'src',
        'dist',
        ['wheel'],
        config_settings=config_settings,
        isolation=False,
        skip_dependency_check=True,
        dependency_constraints_txt=pathlib.Path('constraints.txt'),
        installer='uv',
    )

    assert built == ['demo-1.0.0.tar.gz', 'demo-1.0.0-py3-none-any.whl']
    extractall = tar_open.return_value.__enter__.return_value.extractall
    extractall.assert_called_once()
    assert extractall.call_args.args[0] == 'temp-sdist-dir'
    build_cmd.assert_has_calls(
        [
            unittest.mock.call(
                False, 'src', 'dist', 'sdist', config_settings, True, pathlib.Path('constraints.txt'), 'uv', None
            ),
            unittest.mock.call(
                False,
                os.path.join('temp-sdist-dir', 'demo-1.0.0'),
                'dist',
                'wheel',
                config_settings,
                True,
                pathlib.Path('constraints.txt'),
                'uv',
                None,
            ),
        ]
    )
    rmtree.assert_called_once_with('temp-sdist-dir', ignore_errors=True)


def test_build_no_isolation_check_deps_not_installed(mocker: pytest_mock.MockerFixture, package_test_flit: str) -> None:
    error = mocker.patch('build.__main__._error')
    build_cmd = mocker.patch('build.ProjectBuilder.build', return_value='something')
    mocker.patch('build.ProjectBuilder.check_dependencies', return_value=[('foo>=1.0',)])
    mocker.patch('build._compat.importlib.metadata.distribution', side_effect=_importlib.metadata.PackageNotFoundError)

    build.__main__.build_package(package_test_flit, '.', ['sdist'], isolation=False)

    build_cmd.assert_called_with('sdist', '.', None)
    error.assert_called_once_with(
        f'Unmet dependencies (checked against {sys.executable}):\n\tfoo>=1.0\n\t\twanted: >=1.0\n\t\tfound: not installed'
    )


def test_build_no_isolation_check_deps_version_mismatch(mocker: pytest_mock.MockerFixture, package_test_flit: str) -> None:
    error = mocker.patch('build.__main__._error')
    mocker.patch('build.ProjectBuilder.build', return_value='something')
    mocker.patch('build.ProjectBuilder.check_dependencies', return_value=[('bar>=2.0',)])
    mocker.patch('build._compat.importlib.metadata.distribution', return_value=mocker.MagicMock(version='1.0.0'))

    build.__main__.build_package(package_test_flit, '.', ['sdist'], isolation=False)

    error.assert_called_once_with(
        f'Unmet dependencies (checked against {sys.executable}):\n\tbar>=2.0\n\t\twanted: >=2.0\n\t\tfound: 1.0.0'
    )


def test_build_no_isolation_check_deps_chain_without_specifier(
    mocker: pytest_mock.MockerFixture, package_test_flit: str
) -> None:
    error = mocker.patch('build.__main__._error')
    mocker.patch('build.ProjectBuilder.build', return_value='something')
    mocker.patch('build.ProjectBuilder.check_dependencies', return_value=[('matplotlib>=2.2', 'kiwisolver')])
    mocker.patch('build._compat.importlib.metadata.distribution', side_effect=_importlib.metadata.PackageNotFoundError)

    build.__main__.build_package(package_test_flit, '.', ['sdist'], isolation=False)

    error.assert_called_once_with(
        f'Unmet dependencies (checked against {sys.executable}):'
        '\n\tmatplotlib>=2.2 -> kiwisolver\n\t\twanted: any\n\t\tfound: not installed'
    )


@pytest.mark.parametrize(
    ('cli_args', 'err_msg'),
    [
        (['-Cone=1', '--config-json={"two": 2}'], 'not allowed with argument'),
        (['--config-json={"two": 2'], 'Invalid JSON in --config-json'),
        (['--config-json=[1]'], '--config-json must contain a JSON object'),
    ],
)
def test_config_json_errors(cli_args: list[str], err_msg: str, capsys: pytest.CaptureFixture[str]) -> None:
    with pytest.raises(SystemExit):
        build.__main__.main(cli_args)

    outerr = capsys.readouterr()
    assert err_msg in outerr.out or err_msg in outerr.err


@pytest.mark.isolated
def test_build_raises_build_exception(mocker: pytest_mock.MockerFixture, package_test_flit: str) -> None:
    mocker.patch('build.ProjectBuilder.get_requires_for_build', side_effect=build.BuildException)
    mocker.patch('build.env.DefaultIsolatedEnv.install')

    with pytest.raises(build.BuildException):
        build.__main__.build_package(package_test_flit, '.', ['sdist'])


@pytest.mark.isolated
def test_build_raises_build_backend_exception(mocker: pytest_mock.MockerFixture, package_test_flit: str) -> None:
    mocker.patch('build.ProjectBuilder.get_requires_for_build', side_effect=build.BuildBackendException(Exception('a')))
    mocker.patch('build.env.DefaultIsolatedEnv.install')

    msg = f"Backend operation failed: Exception('a'{',' if sys.version_info < (3, 7) else ''})"
    with pytest.raises(build.BuildBackendException, match=re.escape(msg)):
        build.__main__.build_package(package_test_flit, '.', ['sdist'])


@pytest.mark.network
def test_build_package(tmp_dir: str, package_test_setuptools: str) -> None:
    build.__main__.build_package(package_test_setuptools, tmp_dir, ['sdist', 'wheel'])

    assert sorted(os.listdir(tmp_dir)) == [
        'test_setuptools-1.0.0-py3-none-any.whl',
        'test_setuptools-1.0.0.tar.gz',
    ]


@pytest.mark.network
def test_build_package_via_sdist(tmp_dir: str, package_test_setuptools: str) -> None:
    build.__main__.build_package_via_sdist(package_test_setuptools, tmp_dir, ['wheel'])

    assert sorted(os.listdir(tmp_dir)) == [
        'test_setuptools-1.0.0-py3-none-any.whl',
        'test_setuptools-1.0.0.tar.gz',
    ]


def test_build_package_via_sdist_incomplete_sdist(tmp_dir: str, package_test_cant_build_via_sdist: str) -> None:
    # The backend produces an sdist without a PKG-INFO, so it is rejected up front rather than
    # failing later when the wheel build cannot find its missing source file.
    with pytest.raises(build.BuildException, match='PKG-INFO'):
        build.__main__.build_package_via_sdist(package_test_cant_build_via_sdist, tmp_dir, ['wheel'])


def test_build_package_via_sdist_invalid_distribution(tmp_dir: str, package_test_setuptools: str) -> None:
    with pytest.raises(ValueError, match='Only binary distributions are allowed but sdist was specified'):
        build.__main__.build_package_via_sdist(package_test_setuptools, tmp_dir, ['sdist'])


@pytest.mark.isolated
def test_build_package_with_constraints(
    mocker: pytest_mock.MockerFixture, tmp_path: pathlib.Path, package_test_flit: str
) -> None:
    install = mocker.patch('build.env.DefaultIsolatedEnv.install')

    constraints_txt_path = tmp_path.joinpath('constraints.txt')
    constraints_txt_path.write_text(
        """\
flit-core==12.34
foo==wot
""",
        encoding='utf-8',
    )

    with pytest.raises(build.BuildBackendException, match=re.escape("Backend 'flit_core.buildapi' is not available.")):
        build.__main__.build_package(package_test_flit, tmp_path, ['wheel'], dependency_constraints_txt=constraints_txt_path)

    install.assert_any_call({'flit_core >=2,<4'}, constraints=('flit-core==12.34\nfoo==wot\n',), _fresh=True)


@pytest.mark.isolated
def test_build_package_with_constraints_passes_file_through_unmodified(
    mocker: pytest_mock.MockerFixture, tmp_path: pathlib.Path, package_test_flit: str
) -> None:
    # As produced by e.g. `pip-compile --generate-hashes`: a requirement and its --hash options wrapped onto
    # continuation lines. Regression test for the requirement/hash pair being split apart when re-parsed into
    # individual lines - the file content must reach the installer byte-for-byte instead.
    install = mocker.patch('build.env.DefaultIsolatedEnv.install')

    constraints_text = (
        'flit-core==12.34 \\\n    --hash=sha256:aaaa \\\n    --hash=sha256:bbbb\n    # via test\nfoo==wot \\\n'
        '    --hash=sha256:cccc\n'
    )
    constraints_txt_path = tmp_path.joinpath('constraints.txt')
    constraints_txt_path.write_text(constraints_text, encoding='utf-8')

    with pytest.raises(build.BuildBackendException, match=re.escape("Backend 'flit_core.buildapi' is not available.")):
        build.__main__.build_package(package_test_flit, tmp_path, ['wheel'], dependency_constraints_txt=constraints_txt_path)

    install.assert_any_call({'flit_core >=2,<4'}, constraints=(constraints_text,), _fresh=True)


@pytest.mark.isolated
def test_build_package_with_empty_constraints_txt(
    mocker: pytest_mock.MockerFixture, tmp_path: pathlib.Path, package_test_flit: str
) -> None:
    install = mocker.patch('build.env.DefaultIsolatedEnv.install')

    constraints_txt_path = tmp_path.joinpath('constraints.txt')
    constraints_txt_path.write_text('  \n\n', encoding='utf-8')

    with pytest.raises(build.BuildBackendException, match=re.escape("Backend 'flit_core.buildapi' is not available.")):
        build.__main__.build_package(package_test_flit, tmp_path, ['wheel'], dependency_constraints_txt=constraints_txt_path)

    install.assert_any_call({'flit_core >=2,<4'}, _fresh=True)


@pytest.mark.parametrize(
    ('args', 'output'),
    [
        pytest.param(
            [],
            [
                '* Creating isolated environment: venv+pip...',
                '* Installing packages in isolated environment:',
                '  - setuptools >= 42.0.0',
                '* Getting build dependencies for sdist...',
                '* Building sdist...',
                '* Building wheel from sdist',
                '* Creating isolated environment: venv+pip...',
                '* Installing packages in isolated environment:',
                '  - setuptools >= 42.0.0',
                '* Getting build dependencies for wheel...',
                '* Building wheel...',
                'Successfully built test_setuptools-1.0.0.tar.gz and test_setuptools-1.0.0-py3-none-any.whl',
            ],
            id='via-sdist-isolation',
            marks=[pytest.mark.network, pytest.mark.isolated],
        ),
        pytest.param(
            ['--no-isolation'],
            [
                '* Getting build dependencies for sdist...',
                '* Building sdist...',
                '* Building wheel from sdist',
                '* Getting build dependencies for wheel...',
                '* Building wheel...',
                'Successfully built test_setuptools-1.0.0.tar.gz and test_setuptools-1.0.0-py3-none-any.whl',
            ],
            id='via-sdist-no-isolation',
        ),
        pytest.param(
            ['--wheel'],
            [
                '* Creating isolated environment: venv+pip...',
                '* Installing packages in isolated environment:',
                '  - setuptools >= 42.0.0',
                '* Getting build dependencies for wheel...',
                '* Building wheel...',
                'Successfully built test_setuptools-1.0.0-py3-none-any.whl',
            ],
            id='wheel-direct-isolation',
            marks=[pytest.mark.network, pytest.mark.isolated],
        ),
        pytest.param(
            ['--wheel', '--no-isolation'],
            [
                '* Getting build dependencies for wheel...',
                '* Building wheel...',
                'Successfully built test_setuptools-1.0.0-py3-none-any.whl',
            ],
            id='wheel-direct-no-isolation',
        ),
        pytest.param(
            ['--sdist', '--no-isolation'],
            [
                '* Getting build dependencies for sdist...',
                '* Building sdist...',
                'Successfully built test_setuptools-1.0.0.tar.gz',
            ],
            id='sdist-direct-no-isolation',
        ),
        pytest.param(
            ['--sdist', '--wheel', '--no-isolation'],
            [
                '* Getting build dependencies for sdist...',
                '* Building sdist...',
                '* Getting build dependencies for wheel...',
                '* Building wheel...',
                'Successfully built test_setuptools-1.0.0.tar.gz and test_setuptools-1.0.0-py3-none-any.whl',
            ],
            id='sdist-and-wheel-direct-no-isolation',
        ),
        pytest.param(
            ['--metadata'],
            [
                '* Creating isolated environment: venv+pip...',
                '* Getting metadata for wheel...',
                '* Getting build dependencies for wheel...',
                '* Installing packages in isolated environment:',
                '  - setuptools >= 42.0.0',
            ],
            id='metadata-isolation',
            marks=[pytest.mark.network, pytest.mark.isolated],
        ),
        pytest.param(
            ['--metadata', '--no-isolation'],
            [
                '* Getting build dependencies for wheel...',
                '* Getting metadata for wheel...',
            ],
            id='metadata-no-isolation',
        ),
    ],
)
@pytest.mark.flaky(reruns=5)
def test_logging_output(
    package_test_setuptools: str, tmp_dir: str, capsys: pytest.CaptureFixture[str], args: list[str], output: list[str]
) -> None:
    build.__main__.main([package_test_setuptools, '-o', tmp_dir, *args])
    _, stderr = capsys.readouterr()
    # the installed-version step lists dynamic versions, exercised in its own test below
    version_pin = re.compile(r' {2}- \S+==\S+')
    lines = [
        line
        for line in stderr.splitlines()
        if line != '* Installed build dependency versions:' and not version_pin.fullmatch(line)
    ]
    assert set(lines) <= set(output)


@pytest.mark.network
@pytest.mark.isolated
@pytest.mark.flaky(reruns=5)
def test_logging_output_backend_versions(
    package_test_setuptools: str, tmp_dir: str, capsys: pytest.CaptureFixture[str]
) -> None:
    build.__main__.main([package_test_setuptools, '-o', tmp_dir, '--wheel'])
    _, stderr = capsys.readouterr()
    assert '* Installed build dependency versions:' in stderr.splitlines()
    assert any(re.fullmatch(r' {2}- setuptools==\d[\w.]*', line) for line in stderr.splitlines())


@pytest.mark.parametrize(
    ('color', 'stderr_error', 'stderr_body'),
    [
        (
            False,
            'ERROR ',
            [
                '* Creating isolated environment: venv+pip...',
                '* Installing packages in isolated environment:',
                '  - setuptools >= 42.0.0',
                '  - this is invalid',
            ],
        ),
        (
            True,
            '\33[91mERROR\33[0m ',
            [
                '\33[1m* Creating isolated environment: venv+pip...\33[0m',
                '\33[1m* Installing packages in isolated environment:\33[0m',
                '  - setuptools >= 42.0.0',
                '  - this is invalid',
            ],
        ),
    ],
    ids=['no-color', 'color'],
)
@pytest.mark.usefixtures('local_pip')
def test_logging_output_env_subprocess_error(
    mocker: pytest_mock.MockerFixture,
    monkeypatch: pytest.MonkeyPatch,
    package_test_invalid_requirements: str,
    tmp_dir: str,
    capsys: pytest.CaptureFixture[str],
    color: bool,
    stderr_body: list[str],
    stderr_error: str,
) -> None:
    with contextlib.suppress(ModuleNotFoundError):  # colorama might not be available
        # do not inject hook to have clear output on capsys
        mocker.patch('colorama.init')

    monkeypatch.delenv('NO_COLOR', raising=False)
    monkeypatch.setenv('FORCE_COLOR' if color else 'NO_COLOR', '')

    with pytest.raises(SystemExit):
        build.__main__.main([package_test_invalid_requirements, '-o', tmp_dir])
    outerr = capsys.readouterr()
    stderr = outerr.err.splitlines()

    assert stderr[:4] == stderr_body
    assert stderr[-1].startswith(stderr_error)

    # Newer versions of pip also color stderr - strip them if present
    assert any(ANSI_STRIP.sub('', e).strip().startswith('< ERROR: Invalid requirement: ') for e in stderr)


@pytest.mark.parametrize(
    ('stdout_tty', 'stderr_tty', 'env', 'stdout_colors', 'stderr_colors'),
    [
        (False, False, {}, build.__main__._NO_COLORS, build.__main__._NO_COLORS),
        (True, True, {}, build.__main__._COLORS, build.__main__._COLORS),
        # regression: color support must be decided per-stream, not solely from stdout.
        (False, True, {}, build.__main__._NO_COLORS, build.__main__._COLORS),
        (True, False, {}, build.__main__._COLORS, build.__main__._NO_COLORS),
        (False, False, {'NO_COLOR': ''}, build.__main__._NO_COLORS, build.__main__._NO_COLORS),
        (True, True, {'NO_COLOR': ''}, build.__main__._NO_COLORS, build.__main__._NO_COLORS),
        (False, False, {'FORCE_COLOR': ''}, build.__main__._COLORS, build.__main__._COLORS),
        (True, True, {'FORCE_COLOR': ''}, build.__main__._COLORS, build.__main__._COLORS),
    ],
)
def test_colors(
    mocker: pytest_mock.MockerFixture,
    monkeypatch: pytest.MonkeyPatch,
    stdout_tty: bool,
    stderr_tty: bool,
    env: dict[str, str],
    stdout_colors: dict[str, object],
    stderr_colors: dict[str, object],
) -> None:
    mocker.patch('sys.stdout.isatty', return_value=stdout_tty)
    mocker.patch('sys.stderr.isatty', return_value=stderr_tty)
    monkeypatch.delenv('NO_COLOR', raising=False)
    monkeypatch.delenv('FORCE_COLOR', raising=False)
    for key, value in env.items():
        monkeypatch.setenv(key, value)

    build.__main__._init_colors()

    styles = build.__main__._styles.get()
    assert styles['stdout'] == stdout_colors
    assert styles['stderr'] == stderr_colors


def test_colors_conflict(monkeypatch: pytest.MonkeyPatch) -> None:
    with monkeypatch.context() as m:
        m.setenv('NO_COLOR', '')
        m.setenv('FORCE_COLOR', '')

        with pytest.warns(
            UserWarning,
            match='Both NO_COLOR and FORCE_COLOR environment variables are set, disabling color',
        ):
            build.__main__._init_colors()

        styles = build.__main__._styles.get()
        assert styles['stdout'] == build.__main__._NO_COLORS
        assert styles['stderr'] == build.__main__._NO_COLORS


def test_cprint_selects_style_per_stream(mocker: pytest_mock.MockerFixture, monkeypatch: pytest.MonkeyPatch) -> None:
    """stderr output is colored independently of stdout, based on its own tty-ness."""
    monkeypatch.delenv('NO_COLOR', raising=False)
    monkeypatch.delenv('FORCE_COLOR', raising=False)

    # stdout is not a tty (e.g. redirected to a log file), stderr is a tty.
    mocker.patch('sys.stdout.isatty', return_value=False)
    mocker.patch('sys.stderr.isatty', return_value=True)
    build.__main__._init_colors()

    assert build.__main__._styles.get()['stdout'] == build.__main__._NO_COLORS
    assert build.__main__._styles.get()['stderr'] == build.__main__._COLORS

    # the inverse: stdout is a tty, stderr is redirected.
    mocker.patch('sys.stdout.isatty', return_value=True)
    mocker.patch('sys.stderr.isatty', return_value=False)
    build.__main__._init_colors()

    assert build.__main__._styles.get()['stdout'] == build.__main__._COLORS
    assert build.__main__._styles.get()['stderr'] == build.__main__._NO_COLORS


def test_warning_goes_to_stderr(capsys: pytest.CaptureFixture[str]) -> None:
    build.__main__._showwarning('a warning', UserWarning, 'f.py', 1)

    out, err = capsys.readouterr()

    assert out == ''
    assert 'WARNING' in err
    assert 'a warning' in err


def test_logging_output_venv_failure(
    monkeypatch: pytest.MonkeyPatch, package_test_flit: str, tmp_dir: str, capsys: pytest.CaptureFixture[str]
) -> None:
    def raise_called_process_err(*_args: object, **_kwargs: object) -> None:
        raise subprocess.CalledProcessError(1, ['test', 'args'], b'stdoutput', b'stderror')

    monkeypatch.setattr(venv.EnvBuilder, 'create', raise_called_process_err)
    monkeypatch.setenv('NO_COLOR', '')

    with pytest.raises(SystemExit):
        build.__main__.main([package_test_flit, '-o', tmp_dir])

    _, stderr = capsys.readouterr()

    assert (
        stderr
        == """\
* Creating isolated environment: venv+pip...
> test args
< stdoutput
< stderror
ERROR Failed to create venv. Maybe try installing virtualenv.
"""
    )


@pytest.mark.contextvars
@pytest.mark.network
def test_verbose_logging_output(
    subtests: SubTests,
    capfd: pytest.CaptureFixture[str],
    monkeypatch: pytest.MonkeyPatch,
    tmp_dir: str,
    package_test_setuptools: str,
) -> None:
    monkeypatch.setenv('NO_COLOR', '')

    no_of_lines = -1

    for verbosity in range(-2, 3):
        with subtests.test(verbosity=verbosity):
            cmd = [package_test_setuptools, '-w', '-o', tmp_dir]
            if verbosity:
                cmd.insert(0, f'-{("v" if verbosity > 0 else "q") * abs(verbosity)}')

            build.__main__.main(cmd)

            new_no_of_lines = sum(1 for s in capfd.readouterr() for i in s.splitlines())
            assert new_no_of_lines > no_of_lines
            no_of_lines = new_no_of_lines


def test_metadata_json_output(
    capsys: pytest.CaptureFixture[str],
    package_test_setuptools: str,
) -> None:
    build.__main__.main([package_test_setuptools, '--metadata', '-n'])

    stdout = capsys.readouterr().out
    metadata = json.loads(stdout)
    # Name normalised in old versions of setuptools.
    assert metadata['name'] in {'test_setuptools', 'test-setuptools'}
    assert metadata['version'] == '1.0.0'


def test_setup_cli_windows_colorama_available(mocker: pytest_mock.MockerFixture) -> None:
    mocker.patch('platform.system', return_value='Windows')
    colorama = mocker.MagicMock()
    mocker.patch.dict('sys.modules', {'colorama': colorama})
    build.__main__._setup_cli(verbosity=0)
    colorama.init.assert_called_once()


def test_setup_cli_windows_colorama_missing(mocker: pytest_mock.MockerFixture) -> None:
    mocker.patch('platform.system', return_value='Windows')
    mocker.patch.dict('sys.modules', {'colorama': None})
    build.__main__._setup_cli(verbosity=0)


def test_setup_cli_non_windows(mocker: pytest_mock.MockerFixture) -> None:
    mocker.patch('platform.system', return_value='Linux')
    build.__main__._setup_cli(verbosity=0)


def test_metadata_with_distributions_error() -> None:
    with pytest.raises(SystemExit):
        build.__main__.main(['--metadata', '--sdist'])


def test_entrypoint(mocker: pytest_mock.MockerFixture) -> None:
    main = mocker.patch('build.__main__.main')
    build.__main__.entrypoint()
    main.assert_called_once_with(sys.argv[1:])


def test_handle_build_error_build_backend_exception(mocker: pytest_mock.MockerFixture) -> None:
    mocker.patch('build.__main__._error', side_effect=SystemExit(1))

    with pytest.raises(SystemExit), build.__main__._handle_build_error(env_dir=None, sdist_extract_dir=None):
        raise build.BuildBackendException(
            ValueError('test error'),
        )


def test_handle_build_error_prints_debug_tip_on_subprocess_failure(
    mocker: pytest_mock.MockerFixture,
    capsys: pytest.CaptureFixture[str],
) -> None:
    mocker.patch('build.__main__._error', side_effect=SystemExit(1))

    exc = subprocess.CalledProcessError(1, ['backend'])
    with pytest.raises(SystemExit), build.__main__._handle_build_error(env_dir=None, sdist_extract_dir=None):
        raise build.BuildBackendException(exc, 'Backend subprocess exited')

    err = ANSI_STRIP.sub('', capsys.readouterr().err)
    assert 'TIP' in err
    assert 'pass --env-dir and --sdist-extract-dir' in err
    assert 'troubleshooting.html#debug-a-failed-build' in err


@pytest.mark.parametrize(
    ('env_dir', 'sdist_extract_dir', 'expected'),
    [
        pytest.param(None, None, 'pass --env-dir and --sdist-extract-dir to keep', id='none-pinned'),
        pytest.param('env', None, 'inspect the build environment at env', id='env-only'),
        pytest.param(None, 'src', 'inspect the extracted sources at src', id='sdist-only'),
        pytest.param('env', 'src', 'inspect the build environment at env and the extracted sources at src', id='both'),
    ],
)
def test_build_failure_hint(env_dir: str | None, sdist_extract_dir: str | None, expected: str) -> None:
    hint = build.__main__._build_failure_hint(env_dir, sdist_extract_dir)

    assert expected in hint
    assert hint.endswith('troubleshooting.html#debug-a-failed-build for help debugging a failed build')


def test_log_unknown_kind(mocker: pytest_mock.MockerFixture) -> None:
    mocker.patch('build.__main__._cprint')
    log = build.__main__._make_logger()
    log('message', kind=('unknown',))


def test_log_dependency_versions(mocker: pytest_mock.MockerFixture) -> None:
    env = mocker.create_autospec(build.env.DefaultIsolatedEnv, instance=True)
    env.installed_versions.return_value = {'wheel': '0.45.1', 'setuptools': '80.9.0'}
    log = mocker.patch('build.__main__._ctx.log')

    build.__main__._log_dependency_versions(env, {'setuptools', 'wheel'})

    log.assert_called_once_with(
        'Installed build dependency versions:\n- setuptools==80.9.0\n- wheel==0.45.1',
        kind=('step',),
    )


def test_log_dependency_versions_none(mocker: pytest_mock.MockerFixture) -> None:
    env = mocker.create_autospec(build.env.DefaultIsolatedEnv, instance=True)
    env.installed_versions.return_value = {}
    log = mocker.patch('build.__main__._ctx.log')

    build.__main__._log_dependency_versions(env, set())

    log.assert_not_called()


def test_bootstrap_build_env_logs_versions(mocker: pytest_mock.MockerFixture) -> None:
    env = mocker.create_autospec(build.env.DefaultIsolatedEnv, instance=True)
    env.__enter__.return_value = env
    mocker.patch('build.__main__.DefaultIsolatedEnv', return_value=env)
    builder = mocker.create_autospec(build.ProjectBuilder, instance=True)
    builder.build_system_requires = {'setuptools'}
    builder.get_requires_for_build.return_value = {'wheel'}
    mocker.patch('build.ProjectBuilder.from_isolated_env', return_value=builder)
    log_versions = mocker.patch('build.__main__._log_dependency_versions')

    with build.__main__._bootstrap_build_env(
        isolation=True,
        srcdir='src',
        distribution='wheel',
        config_settings=None,
        skip_dependency_check=False,
        dependency_constraints_txt=None,
        installer='pip',
    ) as result:
        assert result is builder

    env.install.assert_any_call({'setuptools'}, _fresh=True)
    env.install.assert_any_call({'wheel'})
    log_versions.assert_called_once_with(env, {'setuptools', 'wheel'})


def test_build_no_isolation_skip_dependency_check(mocker: pytest_mock.MockerFixture, package_test_flit: str) -> None:
    build_cmd = mocker.patch('build.ProjectBuilder.build', return_value='something')
    build.__main__.build_package(package_test_flit, '.', ['sdist'], isolation=False, skip_dependency_check=True)
    build_cmd.assert_called_with('sdist', '.', None)


def test_build_package_via_sdist_empty_distributions(mocker: pytest_mock.MockerFixture) -> None:
    mocker.patch(
        'build.__main__._build',
        return_value=os.path.join('dist', 'demo-1.0.0.tar.gz'),
    )
    result = build.__main__.build_package_via_sdist('src', 'dist', [])
    assert result == ['demo-1.0.0.tar.gz']


def test_parse_config_settings_triple_duplicate() -> None:
    result = build.__main__._parse_config_settings(['--flag=a', '--flag=b', '--flag=c'])
    assert result == {'--flag': ['a', 'b', 'c']}


@pytest.mark.parametrize(
    ('arg', 'value', 'warns'),
    [
        pytest.param('--flag', '', True, id='bare-key-warns'),
        pytest.param('--flag=', '', False, id='empty-value'),
        pytest.param('--flag=value', 'value', False, id='with-value'),
    ],
)
def test_parse_config_settings_pip_compatibility(arg: str, value: str, warns: bool, recwarn: pytest.WarningsRecorder) -> None:
    assert build.__main__._parse_config_settings([arg]) == {'--flag': value}

    messages = [str(warning.message) for warning in recwarn]
    config_warnings = [m for m in messages if "Config setting '--flag' was passed without a value" in m]
    assert messages == config_warnings  # no unrelated warnings are emitted either way
    assert bool(config_warnings) == warns


def test_build_metadata_runner_without_extra_environ(
    mocker: pytest_mock.MockerFixture,
    tmp_path: pathlib.Path,
    package_test_setuptools: str,
) -> None:
    captured_runners: list[SubprocessRunner] = []

    @contextlib.contextmanager
    def fake_bootstrap(*_args: object, runner: SubprocessRunner, **_kwargs: object) -> Generator[unittest.mock.MagicMock]:
        captured_runners.append(runner)
        builder = mocker.MagicMock()
        metadata_dir = tmp_path / 'metadata'
        metadata_dir.mkdir()
        (metadata_dir / 'METADATA').write_bytes(b'Metadata-Version: 2.2\nName: test\nVersion: 1.0\n')
        builder.metadata_path.return_value = str(metadata_dir)
        yield builder

    mocker.patch('build.__main__._bootstrap_build_env', side_effect=fake_bootstrap)
    ctx_run = mocker.patch('build.__main__._ctx.run_subprocess')

    build.__main__._build_metadata(package_test_setuptools, '.', ['wheel'], isolation=False, skip_dependency_check=True)

    assert captured_runners
    captured_runners[0](['echo', 'test'])
    ctx_run.assert_called_once_with(['echo', 'test'], None, mocker.ANY)


class WriteSdist(Protocol):
    def __call__(
        self, path: pathlib.Path, top_level: str, *, with_pkg_info: bool = ..., extra: dict[str, str] | None = ...
    ) -> None: ...


@pytest.fixture
def write_sdist() -> WriteSdist:
    def _write(
        path: pathlib.Path,
        top_level: str,
        *,
        with_pkg_info: bool = True,
        extra: dict[str, str] | None = None,
    ) -> None:
        pkg_info = b'Metadata-Version: 2.2\nName: demo\nVersion: 1.0.0\n'
        with tarfile.open(path, 'w:gz') as tar:
            if with_pkg_info:
                info = tarfile.TarInfo(name=f'{top_level}/PKG-INFO')
                info.size = len(pkg_info)
                tar.addfile(info, io.BytesIO(pkg_info))
            body = b'[build-system]\nrequires = []\nbuild-backend = "noop"\n'
            info = tarfile.TarInfo(name=f'{top_level}/pyproject.toml')
            info.size = len(body)
            tar.addfile(info, io.BytesIO(body))
            for member_name, member_body in (extra or {}).items():
                data = member_body.encode()
                info = tarfile.TarInfo(name=member_name)
                info.size = len(data)
                tar.addfile(info, io.BytesIO(data))

    return _write


@pytest.fixture
def sdist(tmp_path: pathlib.Path, write_sdist: WriteSdist) -> pathlib.Path:
    archive = tmp_path / 'demo-1.0.0.tar.gz'
    write_sdist(archive, 'demo-1.0.0')
    return archive


def test_validate_sdist_archive_happy(sdist: pathlib.Path) -> None:
    assert build.__main__._validate_sdist_archive(str(sdist)) == 'demo-1.0.0'


def test_validate_sdist_archive_top_level_name_mismatch(tmp_path: pathlib.Path, write_sdist: WriteSdist) -> None:
    archive = tmp_path / 'demo-1.0.0.tar.gz'
    write_sdist(archive, 'something-else-1.0')
    assert build.__main__._validate_sdist_archive(str(archive)) == 'something-else-1.0'


def _invalid_filename(tmp_path: pathlib.Path, write: WriteSdist) -> str:
    archive = tmp_path / 'noversion.tar.gz'
    write(archive, 'noversion')
    return str(archive)


def _invalid_version(tmp_path: pathlib.Path, write: WriteSdist) -> str:
    archive = tmp_path / 'demo-not_a_version.tar.gz'
    write(archive, 'demo-not_a_version')
    return str(archive)


def _corrupt_tar(tmp_path: pathlib.Path, _write: WriteSdist) -> str:
    archive = tmp_path / 'demo-1.0.0.tar.gz'
    archive.write_bytes(b'not a real tar.gz')
    return str(archive)


def _multiple_top_level(tmp_path: pathlib.Path, write: WriteSdist) -> str:
    archive = tmp_path / 'demo-1.0.0.tar.gz'
    write(archive, 'demo-1.0.0', extra={'other-1.0.0/PKG-INFO': 'x'})
    return str(archive)


def _missing_pkg_info(tmp_path: pathlib.Path, write: WriteSdist) -> str:
    archive = tmp_path / 'demo-1.0.0.tar.gz'
    write(archive, 'demo-1.0.0', with_pkg_info=False)
    return str(archive)


_REJECT_CASES: dict[str, Callable[[pathlib.Path, WriteSdist], str]] = {
    'invalid-filename': _invalid_filename,
    'invalid-version': _invalid_version,
    'corrupt-tar': _corrupt_tar,
    'multiple-top-level': _multiple_top_level,
    'missing-pkg-info': _missing_pkg_info,
}


@pytest.mark.parametrize(
    ('case', 'match'),
    [
        pytest.param('invalid-filename', 'does not look like a source distribution', id='invalid-filename'),
        pytest.param('invalid-version', 'does not look like a source distribution', id='invalid-version'),
        pytest.param('corrupt-tar', 'failed to read source distribution', id='corrupt-tar'),
        pytest.param('multiple-top-level', 'single top-level directory', id='multiple-top-level'),
        pytest.param('missing-pkg-info', 'does not contain demo-1.0.0/PKG-INFO', id='missing-pkg-info'),
    ],
)
def test_validate_sdist_archive_rejects(
    tmp_path: pathlib.Path,
    write_sdist: WriteSdist,
    case: str,
    match: str,
) -> None:
    archive = _REJECT_CASES[case](tmp_path, write_sdist)
    with pytest.raises(build.BuildException, match=match):
        build.__main__._validate_sdist_archive(archive)


def test_extract_sdist_yields_top_level(sdist: pathlib.Path) -> None:
    with build.__main__._extract_sdist(str(sdist), 'demo-1.0.0') as extracted:
        assert os.path.isdir(extracted)
        assert os.path.isfile(os.path.join(extracted, 'PKG-INFO'))
        extract_root = os.path.dirname(extracted)
    assert not os.path.exists(extract_root)


def _raise_inside_extract(sdist: pathlib.Path) -> None:
    with build.__main__._extract_sdist(str(sdist), 'demo-1.0.0') as extracted:
        msg = 'boom'
        raise RuntimeError(msg, os.path.dirname(extracted))


def test_extract_sdist_cleans_up_on_error(sdist: pathlib.Path) -> None:
    with pytest.raises(RuntimeError, match='boom') as exc_info:
        _raise_inside_extract(sdist)
    _, extract_root = exc_info.value.args
    assert not os.path.exists(extract_root)


def test_extract_sdist_rejects_path_traversal(tmp_path: pathlib.Path) -> None:
    archive = tmp_path / 'demo-1.0.0.tar.gz'
    body = b'evil\n'
    with tarfile.open(archive, 'w:gz') as tar:
        info = tarfile.TarInfo(name='../evil.txt')
        info.size = len(body)
        tar.addfile(info, io.BytesIO(body))

    cm = build.__main__._extract_sdist(str(archive), 'demo-1.0.0')
    with pytest.raises(tarfile.TarError):
        cm.__enter__()
    assert not (tmp_path / 'evil.txt').exists()


def test_extract_sdist_rejects_absolute_symlink(tmp_path: pathlib.Path) -> None:
    archive = tmp_path / 'demo-1.0.0.tar.gz'
    with tarfile.open(archive, 'w:gz') as tar:
        info = tarfile.TarInfo(name='demo-1.0.0/evil')
        info.type = tarfile.SYMTYPE
        info.linkname = '/etc/passwd'
        tar.addfile(info)

    cm = build.__main__._extract_sdist(str(archive), 'demo-1.0.0')
    with pytest.raises(tarfile.TarError):
        cm.__enter__()


def test_extract_sdist_fixed_dir_is_deterministic_and_kept(sdist: pathlib.Path, tmp_path: pathlib.Path) -> None:
    extract_dir = tmp_path / 'extract'
    with build.__main__._extract_sdist(str(sdist), 'demo-1.0.0', extract_dir=str(extract_dir)) as extracted:
        assert extracted == str(extract_dir / 'demo-1.0.0')
        assert os.path.isfile(os.path.join(extracted, 'PKG-INFO'))
    assert (extract_dir / 'demo-1.0.0' / 'PKG-INFO').is_file()


def test_extract_sdist_fixed_dir_clears_stale_before_extract(sdist: pathlib.Path, tmp_path: pathlib.Path) -> None:
    extract_dir = tmp_path / 'extract'
    stale = extract_dir / 'demo-1.0.0' / 'stale.txt'
    stale.parent.mkdir(parents=True)
    stale.write_text('old', encoding='utf-8')

    with build.__main__._extract_sdist(str(sdist), 'demo-1.0.0', extract_dir=str(extract_dir)) as extracted:
        assert extracted == str(extract_dir / 'demo-1.0.0')
        assert not stale.exists()
        assert os.path.isfile(os.path.join(extracted, 'PKG-INFO'))


@pytest.mark.parametrize(
    ('extra_args', 'expected'),
    [
        pytest.param([], None, id='default-temp-dir'),
        pytest.param(['--sdist-extract-dir', 'extract-here'], 'extract-here', id='fixed-dir'),
    ],
)
def test_main_via_sdist_forwards_extract_dir(
    tmp_path: pathlib.Path,
    mocker: pytest_mock.MockerFixture,
    extra_args: list[str],
    expected: str | None,
) -> None:
    via_sdist = mocker.patch('build.__main__.build_package_via_sdist', autospec=True, return_value=['something'])

    build.__main__.main([str(tmp_path), *extra_args])

    assert via_sdist.call_args.kwargs['sdist_extract_dir'] == expected


def test_main_sdist_input_forwards_extract_dir(
    sdist: pathlib.Path,
    tmp_path: pathlib.Path,
    mocker: pytest_mock.MockerFixture,
) -> None:
    extract = mocker.patch('build.__main__._extract_sdist', autospec=True)
    extract.return_value.__enter__.return_value = str(tmp_path / 'demo-1.0.0')
    mocker.patch('build.__main__.build_package', return_value=['demo-1.0.0-py3-none-any.whl'])

    build.__main__.main([str(sdist), '--wheel', '-o', str(tmp_path), '--sdist-extract-dir', 'extract-here'])

    assert extract.call_args.kwargs['extract_dir'] == 'extract-here'


@pytest.mark.parametrize(
    ('extra_args', 'expected_distributions', 'expected_outdir'),
    [
        pytest.param([], ['wheel'], None, id='no-flag'),
        pytest.param(['--wheel'], ['wheel'], None, id='wheel-long'),
        pytest.param(['-w'], ['wheel'], None, id='wheel-short'),
        pytest.param(['--wheel', '-o', 'custom-out'], ['wheel'], 'custom-out', id='custom-outdir'),
    ],
)
def test_main_sdist_input_wheel_dispatch(
    sdist: pathlib.Path,
    mocker: pytest_mock.MockerFixture,
    extra_args: list[str],
    expected_distributions: list[str],
    expected_outdir: str | None,
) -> None:
    build_package = mocker.patch('build.__main__.build_package', return_value=['demo-1.0.0-py3-none-any.whl'])
    via_sdist = mocker.patch('build.__main__.build_package_via_sdist')

    build.__main__.main([str(sdist), *extra_args])

    via_sdist.assert_not_called()
    args, kwargs = build_package.call_args
    extracted_srcdir, outdir = args
    assert os.path.basename(extracted_srcdir) == 'demo-1.0.0'
    assert kwargs['distributions'] == expected_distributions
    expected = expected_outdir if expected_outdir is not None else os.path.dirname(os.path.abspath(sdist))
    assert outdir == expected


def test_main_sdist_input_metadata(sdist: pathlib.Path, mocker: pytest_mock.MockerFixture) -> None:
    build_metadata = mocker.patch('build.__main__._build_metadata', return_value=[])

    build.__main__.main([str(sdist), '--metadata'])

    args, kwargs = build_metadata.call_args
    extracted_srcdir, _outdir = args
    assert os.path.basename(extracted_srcdir) == 'demo-1.0.0'
    assert kwargs['distributions'] == ['wheel']


@pytest.mark.parametrize(
    'flags',
    [
        pytest.param(['--sdist'], id='sdist-only'),
        pytest.param(['--sdist', '--wheel'], id='sdist-and-wheel'),
    ],
)
def test_main_sdist_input_rejects_sdist_flag(
    sdist: pathlib.Path,
    capsys: pytest.CaptureFixture[str],
    flags: list[str],
) -> None:
    with pytest.raises(SystemExit):
        build.__main__.main([str(sdist), *flags])
    assert 'cannot build a source distribution from a source distribution' in capsys.readouterr().err


def test_main_sdist_input_validation_error_surfaced(
    tmp_path: pathlib.Path,
    write_sdist: WriteSdist,
    capsys: pytest.CaptureFixture[str],
) -> None:
    archive = tmp_path / 'demo-1.0.0.tar.gz'
    write_sdist(archive, 'demo-1.0.0', with_pkg_info=False)

    with pytest.raises(SystemExit):
        build.__main__.main([str(archive), '--wheel'])

    assert 'PKG-INFO' in capsys.readouterr().err


def test_main_sdist_input_passes_kwargs(sdist: pathlib.Path, mocker: pytest_mock.MockerFixture) -> None:
    build_package = mocker.patch('build.__main__.build_package', return_value=['demo-1.0.0-py3-none-any.whl'])

    build.__main__.main([str(sdist), '--wheel', '-n', '-x', '-Cflag=value'])

    _args, kwargs = build_package.call_args
    assert kwargs['isolation'] is False
    assert kwargs['skip_dependency_check'] is True
    assert kwargs['config_settings'] == {'flag': 'value'}
    assert kwargs['installer'] == 'pip'


@pytest.mark.network
def test_main_sdist_input_end_to_end(tmp_path: pathlib.Path, package_test_setuptools: str) -> None:
    sdist_dir = tmp_path / 'dist'
    sdist_dir.mkdir()
    build.__main__.build_package(package_test_setuptools, str(sdist_dir), ['sdist'])
    archive = next(sdist_dir.glob('*.tar.gz'))

    out_dir = tmp_path / 'out'
    build.__main__.main([str(archive), '--wheel', '-n', '-o', str(out_dir)])

    wheels = list(out_dir.glob('*.whl'))
    assert len(wheels) == 1
    assert wheels[0].name.startswith('test_setuptools-1.0.0')


def test_main_sdist_input_default_outdir_is_archive_parent(
    tmp_path: pathlib.Path,
    sdist: pathlib.Path,
    mocker: pytest_mock.MockerFixture,
) -> None:
    build_package = mocker.patch('build.__main__.build_package', return_value=['demo-1.0.0-py3-none-any.whl'])

    build.__main__.main([str(sdist)])

    _, outdir = build_package.call_args.args
    assert outdir == str(tmp_path)


def test_main_non_archive_file_treated_as_directory(tmp_path: pathlib.Path, mocker: pytest_mock.MockerFixture) -> None:
    archive = tmp_path / 'demo-1.0.0.zip'
    archive.write_bytes(b'')
    via_sdist = mocker.patch('build.__main__.build_package_via_sdist', return_value=['something'])

    build.__main__.main([str(archive)])

    via_sdist.assert_called_once()


@pytest.fixture
def built_dist(tmp_path: pathlib.Path) -> tuple[pathlib.Path, list[str]]:
    outdir = tmp_path / 'dist'
    outdir.mkdir()
    names = ['test_pkg-1.0.0.tar.gz', 'test_pkg-1.0.0-py3-none-any.whl']
    for name in names:
        (outdir / name).write_bytes(f'contents of {name}'.encode())
    return outdir, names


def test_report_written(
    mocker: pytest_mock.MockerFixture,
    tmp_path: pathlib.Path,
    built_dist: tuple[pathlib.Path, list[str]],
) -> None:
    outdir, names = built_dist
    mocker.patch('build.__main__.build_package_via_sdist', autospec=True, return_value=names)
    report = tmp_path / 'report.json'

    build.__main__.main([str(tmp_path), '-o', str(outdir), '--report', str(report)])

    payload = json.loads(report.read_text(encoding='utf-8'))
    assert payload['version'] == '1.0'
    assert [artifact['name'] for artifact in payload['artifacts']] == names
    for artifact, name in zip(payload['artifacts'], names, strict=True):
        path = outdir / name
        assert artifact['path'] == os.path.join(str(outdir), name)
        assert artifact['kind'] == ('sdist' if name.endswith('.tar.gz') else 'wheel')
        assert artifact['size'] == path.stat().st_size
        assert artifact['hashes'] == {'sha256': hashlib.sha256(path.read_bytes()).hexdigest()}


def test_report_requires_path(capsys: pytest.CaptureFixture[str]) -> None:
    with pytest.raises(SystemExit):
        build.__main__.main(['--report'])

    assert '--report: expected one argument' in capsys.readouterr().err


def test_write_report_is_atomic(tmp_path: pathlib.Path, built_dist: tuple[pathlib.Path, list[str]]) -> None:
    outdir, names = built_dist
    report = tmp_path / 'report.json'

    build.__main__._write_report(report, str(outdir), names)

    assert not list(tmp_path.glob('*.tmp'))
    assert not list(outdir.glob('*.tmp'))


def test_write_report_cleans_tmp_on_failure(
    mocker: pytest_mock.MockerFixture, tmp_path: pathlib.Path, built_dist: tuple[pathlib.Path, list[str]]
) -> None:
    outdir, names = built_dist
    mocker.patch('build.__main__.os.replace', side_effect=OSError('boom'))

    with pytest.raises(OSError, match='boom'):
        build.__main__._write_report(tmp_path / 'report.json', str(outdir), names)

    assert not list(tmp_path.glob('*.tmp'))


def test_report_not_allowed_with_metadata(capsys: pytest.CaptureFixture[str]) -> None:
    with pytest.raises(SystemExit):
        build.__main__.main(['--report', 'r.json', '--metadata'])

    assert '--report: not allowed with --metadata' in capsys.readouterr().err


@pytest.fixture
def wheel(tmp_path: pathlib.Path) -> pathlib.Path:
    path = tmp_path / 'demo-1.0.0-py3-none-any.whl'
    with zipfile.ZipFile(path, 'w') as archive:
        archive.writestr('demo-1.0.0.dist-info/METADATA', 'Metadata-Version: 2.1\nName: demo\nVersion: 1.0.0\n\n')
    return path


def test_metadata_read_from_wheel(wheel: pathlib.Path, capsys: pytest.CaptureFixture[str]) -> None:
    build.__main__.main([str(wheel), '--metadata'])

    metadata = json.loads(capsys.readouterr().out)
    assert metadata['name'] == 'demo'
    assert metadata['version'] == '1.0.0'


def test_wheel_input_requires_metadata(wheel: pathlib.Path, capsys: pytest.CaptureFixture[str]) -> None:
    with pytest.raises(SystemExit):
        build.__main__.main([str(wheel)])

    assert 'a wheel can only be used with --metadata' in capsys.readouterr().err


def test_metadata_from_wheel_without_dist_info(tmp_path: pathlib.Path, capsys: pytest.CaptureFixture[str]) -> None:
    path = tmp_path / 'broken-1.0.0-py3-none-any.whl'
    with zipfile.ZipFile(path, 'w') as archive:
        archive.writestr('demo/__init__.py', '')

    with pytest.raises(SystemExit):
        build.__main__.main([str(path), '--metadata'])

    assert 'is not a valid wheel' in capsys.readouterr().err



from __future__ import annotations


__lazy_modules__ = {'pathlib', 'tarfile'}

import sys
import tarfile

from pathlib import Path


# Per https://peps.python.org/pep-0706/, the "data" filter became the default
# in Python 3.14. The first series of releases with the filter had a broken
# filter that could not process symlinks correctly, so the patch releases that
# fixed it are the lower bounds here.
_HAS_DATA_FILTER = (
    (3, 10, 13) <= sys.version_info < (3, 11) or (3, 11, 5) <= sys.version_info < (3, 12) or sys.version_info >= (3, 12)
)


# On runtimes that ship the stdlib ``data`` filter we delegate to it; the fallback branch is
# only reached on 3.10.0-3.10.12 / 3.11.0-3.11.4 and validates each member manually before extraction.
if _HAS_DATA_FILTER:

    def safe_extractall(tar: tarfile.TarFile, path: Path | str) -> None:  # pragma: no cover
        """Extract every member of ``tar`` into ``path`` via the PEP 706 ``data`` filter."""
        tar.extractall(path, filter='data')

else:

    def safe_extractall(tar: tarfile.TarFile, path: Path | str) -> None:  # pragma: no cover
        """Validate every member of ``tar``, then extract into ``path``.

        Reached on 3.10.0-3.10.12 / 3.11.0-3.11.4 where the stdlib ``data`` filter is missing. Device or special files,
        paths that escape ``path``, and symlinks/hardlinks whose targets resolve outside ``path`` are rejected before
        any write hits the disk.

        """
        base = Path(path).resolve()
        for member in tar.getmembers():
            _validate_safe_member(member, base)
        tar.extractall(path)  # noqa: S202


def _validate_safe_member(member: tarfile.TarInfo, base: Path) -> None:
    if member.ischr() or member.isblk() or member.isfifo():
        msg = f'refusing to extract special device file {member.name!r}'
        raise tarfile.TarError(msg)
    target = (base / member.name).resolve(strict=False)
    if not target.is_relative_to(base):
        msg = f'refusing to extract {member.name!r}: path escapes destination'
        raise tarfile.TarError(msg)
    if member.issym() or member.islnk():
        link_base = target.parent if member.issym() else base
        link_target = (link_base / member.linkname).resolve(strict=False)
        if not link_target.is_relative_to(base):
            msg = f'refusing to extract {member.name!r}: link target escapes destination'
            raise tarfile.TarError(msg)


__all__ = [
    'safe_extractall',
]



# SPDX-License-Identifier: MIT

from __future__ import annotations


__lazy_modules__ = {
    'argparse',
    'build._compat',
    'build._compat.tarfile',
    'build._exceptions',
    'build._util',
    'build.env',
    'functools',
    'hashlib',
    'json',
    'packaging',
    'packaging.utils',
    'packaging.version',
    'platform',
    'pyproject_hooks',
    'shutil',
    'subprocess',
    'tarfile',
    'tempfile',
    'textwrap',
    'traceback',
    'warnings',
    'zipfile',
}

import argparse
import contextlib
import contextvars
import hashlib
import json
import os
import platform
import shutil
import subprocess
import sys
import tempfile
import textwrap
import traceback
import warnings
import zipfile

from functools import partial
from tarfile import TarError
from tarfile import open as tar_open
from typing import NoReturn, TextIO, TypedDict

import pyproject_hooks

from packaging.utils import InvalidSdistFilename, parse_sdist_filename
from packaging.version import InvalidVersion

import build
import build.env as _env

from build import ProjectBuilder, _ctx
from build._compat.tarfile import safe_extractall
from build._exceptions import BuildBackendException, BuildException, FailedProcessError
from build._util import format_unmet_dependencies
from build.env import DefaultIsolatedEnv


TYPE_CHECKING = False

if TYPE_CHECKING:
    from collections.abc import Generator, Mapping, Sequence

    from build._types import ConfigSettings, Distribution, JSONValue, StrPath, SubprocessRunner


class _CliArgs:
    """The arguments :func:`main_parser` produces, typed for the rest of the module."""

    srcdir: str
    verbosity: int
    outdir: str | None
    distributions: list[Distribution] | None
    metadata: bool
    config_settings: list[str] | None
    config_json: str | None
    installer: _env.Installer
    no_isolation: bool
    dependency_constraints_txt: str | None
    skip_dependency_check: bool
    sdist_extract_dir: str | None
    env_dir: str | None
    report: str | None


_COLORS = {
    'red': '\33[91m',
    'green': '\33[92m',
    'yellow': '\33[93m',
    'bold': '\33[1m',
    'dim': '\33[2m',
    'underline': '\33[4m',
    'reset': '\33[0m',
}
_NO_COLORS = dict.fromkeys(_COLORS, '')


_DEFAULT_STYLES = {'stdout': _COLORS, 'stderr': _COLORS}
_styles = contextvars.ContextVar('_styles', default=_DEFAULT_STYLES)


def _init_colors() -> None:
    match os.environ:
        case {'NO_COLOR': _, 'FORCE_COLOR': _}:
            warnings.warn('Both NO_COLOR and FORCE_COLOR environment variables are set, disabling color', stacklevel=2)
            _styles.set({'stdout': _NO_COLORS, 'stderr': _NO_COLORS})
        case {'NO_COLOR': _}:
            _styles.set({'stdout': _NO_COLORS, 'stderr': _NO_COLORS})
        case {'FORCE_COLOR': _}:
            _styles.set({'stdout': _COLORS, 'stderr': _COLORS})
        case _:
            _styles.set(
                {
                    'stdout': _COLORS if sys.stdout.isatty() else _NO_COLORS,
                    'stderr': _COLORS if sys.stderr.isatty() else _NO_COLORS,
                }
            )


def _cprint(fmt: str = '', msg: str = '', file: TextIO | None = None) -> None:
    stream = file or sys.stdout
    key = 'stderr' if stream is sys.stderr else 'stdout'
    print(fmt.format(msg, **_styles.get()[key]), file=file, flush=True)


def _showwarning(
    message: Warning | str,
    category: type[Warning],  # noqa: ARG001
    filename: str,  # noqa: ARG001
    lineno: int,  # noqa: ARG001
    file: TextIO | None = None,
    line: str | None = None,  # noqa: ARG001
) -> None:
    _cprint('{yellow}WARNING{reset} {}', str(message), file if file is not None else sys.stderr)


def _make_logger() -> _ctx.Logger:
    max_terminal_width = shutil.get_terminal_size().columns - 2
    if max_terminal_width <= 0:  # pragma: no cover
        max_terminal_width = 78

    fill = partial(textwrap.fill, subsequent_indent='  ', width=max_terminal_width)

    def emit(message: str, *, indent: str, style: str = '{}') -> None:
        for line in message.splitlines():
            _cprint(style, fill(line, initial_indent=indent), file=sys.stderr)

    # Nested so it cannot be called directly, bypassing the logger setup above.
    def log(message: str, *, kind: tuple[str, ...] | None = None) -> None:
        if _ctx.verbosity < -1:
            return
        match kind:
            case ('step', *_):
                first, _, rest = message.partition('\n')
                _cprint('{bold}{}{reset}', fill(first, initial_indent='* '), file=sys.stderr)
                emit(rest, indent='  ')
            case ('subprocess', 'cmd'):
                emit(message, indent='> ', style='{dim}{}{reset}')
            case ('subprocess', 'stdout' | 'stderr'):
                emit(message, indent='< ', style='{dim}{}{reset}')
            case _:
                emit(message, indent='  ')

    return log


def _setup_cli(*, verbosity: int) -> None:
    warnings.showwarning = _showwarning

    if platform.system() == 'Windows':
        try:
            import colorama

            colorama.init()
        except ModuleNotFoundError:
            pass

    _init_colors()

    _ctx.LOGGER.set(_make_logger())
    _ctx.VERBOSITY.set(verbosity)


def _error(msg: str, code: int = 1) -> NoReturn:  # pragma: no cover
    """
    Print an error message and exit. Will color the output when writing to a TTY.

    :param msg: Error message
    :param code: Error code
    """
    _cprint('{red}ERROR{reset} {}', msg, file=sys.stderr)
    raise SystemExit(code)


@contextlib.contextmanager
def _bootstrap_build_env(
    isolation: bool,
    srcdir: StrPath,
    distribution: Distribution,
    config_settings: ConfigSettings | None,
    skip_dependency_check: bool,
    dependency_constraints_txt: os.PathLike[str] | None,
    installer: _env.Installer,
    env_dir: str | None = None,
    runner: SubprocessRunner | None = None,
) -> Generator[ProjectBuilder]:
    runner = runner or pyproject_hooks.default_subprocess_runner
    if isolation:
        with DefaultIsolatedEnv(installer=installer, path=env_dir) as env:
            builder = ProjectBuilder.from_isolated_env(env, srcdir, runner=runner)

            install = env.install
            if dependency_constraints_txt:
                with open(dependency_constraints_txt, encoding='utf-8') as dependency_constraints_file:
                    constraints_text = dependency_constraints_file.read()
                if constraints_text.strip():
                    # Passed through as a single element instead of re-parsed into lines, so a requirement can never
                    # be split from its `--hash` continuation lines (as produced by `pip-compile --generate-hashes`)
                    # and silently lose its hash check. env.py's installer backends reconstruct the original file via
                    # `'\n'.join(constraints)` (see `_PipInstaller`/`_UvInstaller.install_dependencies`), a no-op here
                    # since there is only one element.
                    install = partial(install, constraints=(constraints_text,))

            # first install the build dependencies
            install(builder.build_system_requires, _fresh=True)
            # then get the extra required dependencies from the backend (which was installed in the call above :P)
            extra_requires = builder.get_requires_for_build(distribution, config_settings)
            install(extra_requires)

            _log_dependency_versions(env, builder.build_system_requires | extra_requires)

            yield builder

    else:
        builder = ProjectBuilder(srcdir, runner=runner)

        if not skip_dependency_check and (missing := builder.check_dependencies(distribution, config_settings)):
            _cprint()
            _error(format_unmet_dependencies(missing))

        yield builder


def _log_dependency_versions(env: DefaultIsolatedEnv, requirements: set[str]) -> None:
    if versions := env.installed_versions(requirements):
        _ctx.log(
            'Installed build dependency versions:\n'
            + '\n'.join(f'- {name}=={version}' for name, version in sorted(versions.items())),
            kind=('step',),
        )


def _build(
    isolation: bool,
    srcdir: StrPath,
    outdir: StrPath,
    distribution: Distribution,
    config_settings: ConfigSettings | None,
    skip_dependency_check: bool,
    dependency_constraints_txt: os.PathLike[str] | None,
    installer: _env.Installer,
    env_dir: str | None = None,
) -> str:
    with _bootstrap_build_env(
        isolation,
        srcdir,
        distribution,
        config_settings,
        skip_dependency_check,
        dependency_constraints_txt,
        installer,
        env_dir,
        pyproject_hooks.quiet_subprocess_runner if _ctx.verbosity < 0 else None,
    ) as builder:
        return builder.build(distribution, outdir, config_settings)


@contextlib.contextmanager
def _handle_build_error(*, env_dir: str | None, sdist_extract_dir: StrPath | None) -> Generator[None]:
    try:
        yield
    except (BuildException, FailedProcessError) as e:
        _error(str(e))
    except BuildBackendException as e:
        hint = _build_failure_hint(env_dir, sdist_extract_dir)
        if isinstance(e.exception, subprocess.CalledProcessError):
            _cprint()
            _cprint('{yellow}TIP{reset} {}', hint, file=sys.stderr)
            _error(str(e))

        inner_exception = e.exception
        tb_lines = traceback.format_exception(inner_exception, limit=-1)
        tb = ''.join(tb_lines)
        _cprint('\n{dim}{}{reset}\n', tb.strip('\n'))
        _cprint('{yellow}TIP{reset} {}', hint, file=sys.stderr)
        _error(str(e))
    except Exception as e:  # pragma: no cover
        tb = traceback.format_exc().strip('\n')
        _cprint('\n{dim}{}{reset}\n', tb)
        _error(str(e))


def _build_failure_hint(env_dir: str | None, sdist_extract_dir: StrPath | None) -> str:
    kept = [
        f'{label} at {os.fspath(path)}'
        for path, label in ((env_dir, 'the build environment'), (sdist_extract_dir, 'the extracted sources'))
        if path is not None
    ]
    if kept:
        action = f'inspect {" and ".join(kept)}'
    else:
        action = 'pass --env-dir and --sdist-extract-dir to keep the build environment and sources'
    docs = 'https://build.pypa.io/en/stable/how-to/troubleshooting.html#debug-a-failed-build'
    return f'{action}, then see {docs} for help debugging a failed build'


def _natural_language_list(elements: Sequence[str]) -> str:
    if len(elements) == 0:
        msg = 'no elements'
        raise IndexError(msg)
    if len(elements) == 1:
        return elements[0]
    return '{} and {}'.format(
        ', '.join(elements[:-1]),
        elements[-1],
    )


def build_package(
    srcdir: StrPath,
    outdir: StrPath,
    distributions: Sequence[Distribution],
    config_settings: ConfigSettings | None = None,
    isolation: bool = True,
    skip_dependency_check: bool = False,
    dependency_constraints_txt: os.PathLike[str] | None = None,
    installer: _env.Installer = 'pip',
    env_dir: str | None = None,
) -> list[str]:
    """
    Run the build process.

    :param srcdir: Source directory
    :param outdir: Output directory
    :param distribution: Distribution to build (sdist or wheel)
    :param config_settings: Configuration settings to be passed to the backend
    :param isolation: Isolate the build in a separate environment
    :param skip_dependency_check: Do not perform the dependency check
    :param env_dir: Location of the isolated build environment (a temporary directory is used when not set)
    """
    built: list[str] = []
    for distribution in distributions:
        out = _build(
            isolation,
            srcdir,
            outdir,
            distribution,
            config_settings,
            skip_dependency_check,
            dependency_constraints_txt,
            installer,
            env_dir,
        )
        built.append(os.path.basename(out))
    return built


def build_package_via_sdist(
    srcdir: StrPath,
    outdir: StrPath,
    distributions: Sequence[Distribution],
    config_settings: ConfigSettings | None = None,
    isolation: bool = True,
    skip_dependency_check: bool = False,
    dependency_constraints_txt: os.PathLike[str] | None = None,
    installer: _env.Installer = 'pip',
    sdist_extract_dir: StrPath | None = None,
    env_dir: str | None = None,
) -> list[str]:
    """
    Build a sdist and then the specified distributions from it.

    :param srcdir: Source directory
    :param outdir: Output directory
    :param distribution: Distribution to build (only wheel)
    :param config_settings: Configuration settings to be passed to the backend
    :param isolation: Isolate the build in a separate environment
    :param skip_dependency_check: Do not perform the dependency check
    :param sdist_extract_dir: Directory to extract the intermediate sdist into; a temporary directory is used and
        removed afterwards when ``None``
    :param env_dir: Location of the isolated build environment (a temporary directory is used when not set)
    """
    if 'sdist' in distributions:
        msg = 'Only binary distributions are allowed but sdist was specified'
        raise ValueError(msg)

    sdist = _build(
        isolation,
        srcdir,
        outdir,
        'sdist',
        config_settings,
        skip_dependency_check,
        dependency_constraints_txt,
        installer,
        env_dir,
    )

    sdist_name = os.path.basename(sdist)
    built: list[str] = []
    if distributions:
        top_level = _validate_sdist_archive(sdist)
        with _extract_sdist(sdist, top_level, extract_dir=sdist_extract_dir) as extracted_srcdir:
            _ctx.log(f'Building {_natural_language_list(distributions)} from sdist', kind=('step',))
            for distribution in distributions:
                out = _build(
                    isolation,
                    extracted_srcdir,
                    outdir,
                    distribution,
                    config_settings,
                    skip_dependency_check,
                    dependency_constraints_txt,
                    installer,
                    env_dir,
                )
                built.append(os.path.basename(out))
    return [sdist_name, *built]


def _build_metadata(
    srcdir: StrPath,
    outdir: StrPath,  # noqa: ARG001
    distributions: Sequence[Distribution],  # noqa: ARG001
    config_settings: ConfigSettings | None = None,
    isolation: bool = True,
    skip_dependency_check: bool = False,
    dependency_constraints_txt: os.PathLike[str] | None = None,
    installer: _env.Installer = 'pip',
    env_dir: str | None = None,
) -> list[str]:
    if os.path.isfile(srcdir) and os.fspath(srcdir).lower().endswith('.whl'):
        _print_metadata(_wheel_metadata(srcdir))
        return []

    def run_subprocess(cmd: Sequence[StrPath], cwd: str | None = None, extra_environ: Mapping[str, str] | None = None) -> None:
        env = os.environ.copy()
        if extra_environ:
            env.update(extra_environ)
        _ctx.run_subprocess(cmd, cwd, env)

    with (
        _bootstrap_build_env(
            isolation,
            srcdir,
            'wheel',
            config_settings,
            skip_dependency_check,
            dependency_constraints_txt,
            installer,
            env_dir,
            runner=run_subprocess,
        ) as builder,
        tempfile.TemporaryDirectory() as tempdir,
        open(
            os.path.join(builder.metadata_path(tempdir), 'METADATA'),
            'rb',
        ) as metadata_file,
    ):
        _print_metadata(metadata_file.read())

    return []


def _wheel_metadata(wheel: StrPath) -> bytes:
    with zipfile.ZipFile(wheel) as archive:
        names = [name for name in archive.namelist() if name.count('/') == 1 and name.endswith('.dist-info/METADATA')]
        if len(names) != 1:
            msg = f'{os.fspath(wheel)!r} is not a valid wheel: expected one .dist-info/METADATA, found {len(names)}'
            raise BuildException(msg)
        return archive.read(names[0])


def _print_metadata(raw_metadata: bytes) -> None:
    import packaging.metadata

    valid_metadata, _ = packaging.metadata.parse_email(raw_metadata)
    print(  # noqa: T201
        json.dumps(valid_metadata, ensure_ascii=False, indent=2),
    )


def main_parser() -> argparse.ArgumentParser:
    """
    Construct the main parser.
    """

    class NegativeCountAction(argparse.Action):
        def __init__(
            self,
            option_strings: Sequence[str],
            dest: str,
            default: int | None = None,
            help: str | None = None,
        ) -> None:
            super().__init__(
                option_strings=option_strings,
                dest=dest,
                nargs=0,
                default=default,
                help=help,
            )

        def __call__(
            self,
            parser: argparse.ArgumentParser,  # noqa: ARG002
            namespace: argparse.Namespace,
            values: str | Sequence[str] | None,  # noqa: ARG002
            option_string: str | None = None,  # noqa: ARG002
        ) -> None:
            setattr(namespace, self.dest, getattr(namespace, self.dest, 0) - 1)

    make_parser = partial(
        argparse.ArgumentParser,
        add_help=False,
        description="""\
    A simple, correct Python build frontend.

    By default, a source distribution (sdist) is built from the project root
    and a binary distribution (wheel) is built from the sdist.
    If this is undesirable, you can pass `--sdist` and/or `--wheel`
    to build distributions independently of each other.
    """.rstrip(),
        # Prevent argparse from taking up the entire width of the terminal window
        # which impedes readability. Also keep the description formatted.
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )
    if sys.version_info >= (3, 14):  # pragma: no branch
        make_parser = partial(make_parser, suggest_on_error=True)  # pragma: no cover

    parser = make_parser()
    parser.add_argument(
        'srcdir',
        type=str,
        nargs='?',
        default=os.getcwd(),
        help='source directory, .tar.gz source distribution, or (with --metadata) a .whl to read metadata from '
        '(defaults to the current working directory)',
    )

    global_group = parser.add_argument_group('global options')
    global_group.add_argument(
        '-h',
        '--help',
        action='help',
        default=argparse.SUPPRESS,
        help='show this help message and exit',
    )
    global_group.add_argument(
        '--version',
        '-V',
        action='version',
        version=f'build {build.__version__} ({",".join(build.__path__)})',
    )
    verbosity_exclusive_group = global_group.add_mutually_exclusive_group()
    verbosity_exclusive_group.add_argument(
        '--quiet',
        '-q',
        dest='verbosity',
        action=NegativeCountAction,
        default=0,
        help='reduce verbosity',
    )
    verbosity_exclusive_group.add_argument(
        '--verbose',
        '-v',
        dest='verbosity',
        action='count',
        default=0,
        help='increase verbosity',
    )

    build_group = parser.add_argument_group('build options')
    build_group.add_argument(
        '--outdir',
        '-o',
        help=f'output directory (defaults to {{srcdir}}{os.sep}dist).  Cannot be used together with ``--metadata``',
        metavar='PATH',
    )
    build_group.add_argument(
        '--sdist-extract-dir',
        help='extract the intermediate sdist to PATH (created if missing and kept afterwards) instead of a random '
        'temporary directory; reusing it across rebuilds gives compiler caches such as ccache/sccache a stable '
        'source path. Only affects the default (via-sdist) build and building a wheel from an sdist',
        metavar='PATH',
    )
    build_group.add_argument(
        '--sdist',
        '-s',
        dest='distributions',
        action='append_const',
        const='sdist',
        help='build a source distribution (disables the default behavior)',
    )
    build_group.add_argument(
        '--wheel',
        '-w',
        dest='distributions',
        action='append_const',
        const='wheel',
        help='build a wheel (disables the default behavior)',
    )
    build_group.add_argument(
        '--metadata',
        action='store_true',
        help="print out a wheel's metadata in JSON format, building it first unless the source argument is already a "
        '.whl. Cannot be used in conjunction with ``--sdist`` or ``--wheel``',
    )
    build_group.add_argument(
        '--report',
        help='write a machine-readable JSON report of the built artifacts (name, path, kind, size and SHA-256 hash) '
        'to this path. Cannot be used together with ``--metadata``',
        metavar='PATH',
    )
    config_exclusive_group = build_group.add_mutually_exclusive_group()
    config_exclusive_group.add_argument(
        '--config-setting',
        '-C',
        dest='config_settings',
        action='append',
        help='settings to pass to the backend.  Multiple settings can be provided. '
        'Settings beginning with a hyphen will erroneously be interpreted as options to build if separated '
        'by a space; use ``--config-setting=--my-setting -C--my-other-setting`` instead. '
        'A setting passed without ``=VALUE`` is given an empty value, but this form is not supported by pip; '
        'write ``KEY=`` instead for compatibility',
        metavar='KEY=[VALUE]',
    )
    config_exclusive_group.add_argument(
        '--config-json',
        help='settings to pass to the backend as a JSON object. '
        'This is an alternative to ``--config-setting`` that allows complex nested structures. '
        'Cannot be used together with ``--config-setting``',
        metavar='JSON_STRING',
    )

    install_group = parser.add_argument_group('installation options')
    env_exclusive_group = install_group.add_mutually_exclusive_group()
    env_exclusive_group.add_argument(
        '--installer',
        choices=_env.INSTALLERS,
        default='pip',
        help='Python package installer to use (defaults to pip)',
    )
    env_exclusive_group.add_argument(
        '--no-isolation',
        '-n',
        action='store_true',
        help='disable building the project in an isolated virtual environment. '
        'Build dependencies must be installed separately when this option is used',
    )
    install_group.add_argument(
        '--env-dir',
        help='create the isolated build environment at this location instead of a temporary directory. The location '
        'must be empty; it is removed on success and kept on failure so it can be inspected',
        metavar='PATH',
    )
    install_group.add_argument(
        '--dependency-constraints-txt',
        help='constrain build dependencies using a constraints.txt when installing dependencies',
        metavar='PATH',
    )
    install_group.add_argument(
        '--skip-dependency-check',
        '-x',
        action='store_true',
        help='do not check that build dependencies are installed',
    )

    return parser


def _parse_config_settings(raw_config_settings: list[str]) -> dict[str, str | list[str]]:
    config_settings = dict[str, str | list[str]]()

    for arg in raw_config_settings:
        setting, sep, value = arg.partition('=')
        if not sep:
            warnings.warn(
                f'Config setting {setting!r} was passed without a value; this form is not supported by pip. '
                f'Write {setting}= instead to be compatible with both tools',
                stacklevel=2,
            )
        existing = config_settings.get(setting)
        if existing is None:
            config_settings[setting] = value
        elif isinstance(existing, list):
            existing.append(value)
        else:
            config_settings[setting] = [existing, value]

    return config_settings


def main(cli_args: Sequence[str], prog: str | None = None) -> None:
    """
    Parse the CLI arguments and invoke the build process.

    :param cli_args: CLI arguments
    :param prog: Program name to show in help text
    """
    parser = main_parser()
    if prog:
        parser.prog = prog
    args = parser.parse_args(cli_args, _CliArgs())

    if args.env_dir is not None and args.no_isolation:
        parser.error('--env-dir: not allowed with --no-isolation')

    _setup_cli(verbosity=args.verbosity)

    is_file = os.path.isfile(args.srcdir)
    sdist_input = is_file and os.fspath(args.srcdir).lower().endswith('.tar.gz')
    wheel_input = is_file and os.fspath(args.srcdir).lower().endswith('.whl')
    run_build = partial(
        _select_build(parser, args, sdist_input=sdist_input, wheel_input=wheel_input),
        config_settings=_resolve_config_settings(args),
        isolation=not args.no_isolation,
        skip_dependency_check=args.skip_dependency_check,
        dependency_constraints_txt=args.dependency_constraints_txt,
        installer=args.installer,
        env_dir=args.env_dir,
    )

    if args.outdir is not None:
        outdir = args.outdir
    elif sdist_input or wheel_input:
        outdir = os.path.dirname(os.path.abspath(args.srcdir))
    else:
        outdir = os.path.join(args.srcdir, 'dist')

    with _handle_build_error(env_dir=args.env_dir, sdist_extract_dir=args.sdist_extract_dir):
        if sdist_input:
            top_level = _validate_sdist_archive(args.srcdir)
            with _extract_sdist(args.srcdir, top_level, extract_dir=args.sdist_extract_dir) as extracted_srcdir:
                built = run_build(extracted_srcdir, outdir)
        else:
            built = run_build(args.srcdir, outdir)
        if args.report is not None:
            _write_report(args.report, outdir, built)
        if _ctx.verbosity >= -1 and built:
            artifact_list = _natural_language_list(
                ['{underline}{}{reset}{bold}{green}'.format(artifact, **_styles.get()['stdout']) for artifact in built]
            )
            _cprint('{bold}{green}Successfully built {}{reset}', artifact_list)


class _ArtifactReport(TypedDict):
    name: str
    path: str
    kind: str
    size: int
    hashes: dict[str, str]


class _BuildReport(TypedDict):
    version: str
    artifacts: list[_ArtifactReport]


def _write_report(path: StrPath, outdir: StrPath, artifacts: Sequence[str]) -> None:
    report: _BuildReport = {
        'version': '1.0',
        'artifacts': [_describe_artifact(os.path.join(outdir, name), name) for name in artifacts],
    }
    data = json.dumps(report, indent=2) + '\n'

    fd, tmp = tempfile.mkstemp(dir=os.path.dirname(os.path.abspath(path)), suffix='.tmp')
    try:
        with os.fdopen(fd, 'w', encoding='utf-8') as report_file:
            report_file.write(data)
        os.replace(tmp, path)
    except BaseException:
        with contextlib.suppress(OSError):
            os.unlink(tmp)
        raise


def _describe_artifact(path: str, name: str) -> _ArtifactReport:
    digest = hashlib.sha256()
    size = 0
    with open(path, 'rb') as artifact:
        while chunk := artifact.read(65536):
            size += len(chunk)
            digest.update(chunk)
    return {
        'name': name,
        'path': path,
        'kind': 'sdist' if name.endswith('.tar.gz') else 'wheel',
        'size': size,
        'hashes': {'sha256': digest.hexdigest()},
    }


def _resolve_config_settings(args: _CliArgs) -> Mapping[str, JSONValue]:
    if args.config_json:
        try:
            config_settings = json.loads(args.config_json)
        except json.JSONDecodeError as e:
            _error(f'Invalid JSON in --config-json: {e}')
        if not isinstance(config_settings, dict):
            _error('--config-json must contain a JSON object (dict), not a list or primitive value')
        return config_settings
    if args.config_settings:
        return _parse_config_settings(args.config_settings)
    return {}


def _select_build(
    parser: argparse.ArgumentParser, args: _CliArgs, *, sdist_input: bool, wheel_input: bool
) -> partial[list[str]]:
    if args.report is not None and args.metadata:
        parser.error('--report: not allowed with --metadata')
    if wheel_input and not args.metadata:
        parser.error('a wheel can only be used with --metadata, to read its metadata; it cannot be built from')
    if args.metadata and args.distributions:
        parser.error('--metadata: not allowed with --sdist or --wheel')
    if sdist_input and args.distributions and 'sdist' in args.distributions:
        parser.error(
            'cannot build a source distribution from a source distribution; '
            'pass --wheel to build a wheel from the sdist (see https://github.com/pypa/build/issues/311)'
        )
    if args.metadata:
        return partial(_build_metadata, distributions=['wheel'])
    if sdist_input:
        distributions: list[Distribution] = args.distributions or ['wheel']
        return partial(build_package, distributions=distributions)
    if args.distributions:
        return partial(build_package, distributions=args.distributions)
    return partial(build_package_via_sdist, distributions=['wheel'], sdist_extract_dir=args.sdist_extract_dir)


def _validate_sdist_archive(archive: StrPath) -> str:
    """Validate that ``archive`` is a PEP 625 source distribution and return its top-level directory name."""
    name = os.path.basename(os.fspath(archive))
    try:
        parse_sdist_filename(name)
    except (InvalidSdistFilename, InvalidVersion) as exc:
        msg = f'{name!r} does not look like a source distribution: {exc}'
        raise BuildException(msg) from exc

    try:
        with tar_open(archive) as tar:
            members = tar.getmembers()
    except (OSError, TarError) as exc:
        msg = f'failed to read source distribution {archive}: {exc}'
        raise BuildException(msg) from exc

    top_levels = {m.name.split('/', 1)[0] for m in members if m.name}
    top_levels.discard('')
    if len(top_levels) != 1:
        msg = f'source distribution {archive} must contain a single top-level directory, got: {sorted(top_levels)}'
        raise BuildException(msg)

    top = next(iter(top_levels))
    if not any(m.name == f'{top}/PKG-INFO' and m.isfile() for m in members):
        msg = (
            f'source distribution {archive} does not contain {top}/PKG-INFO; '
            'this does not appear to be a valid source distribution'
        )
        raise BuildException(msg)
    return top


@contextlib.contextmanager
def _extract_sdist(archive: StrPath, top_level: str, *, extract_dir: StrPath | None = None) -> Generator[str]:
    if extract_dir is None:
        tmp_dir = tempfile.mkdtemp(prefix='build-via-sdist-')
        try:
            with tar_open(archive) as tar:
                safe_extractall(tar, tmp_dir)
            yield os.path.join(tmp_dir, top_level)
        finally:
            shutil.rmtree(tmp_dir, ignore_errors=True)
    else:
        root = os.fspath(extract_dir)
        os.makedirs(root, exist_ok=True)
        target = os.path.join(root, top_level)
        shutil.rmtree(target, ignore_errors=True)
        with tar_open(archive) as tar:
            safe_extractall(tar, root)
        yield target


def entrypoint() -> None:
    main(sys.argv[1:])


if __name__ == '__main__':  # pragma: no cover
    main(sys.argv[1:], 'python -m build')


__all__ = [
    'main',
    'main_parser',
]



# SPDX-License-Identifier: MIT

from __future__ import annotations

from collections.abc import Callable
from io import BytesIO
from pathlib import Path
from tarfile import CHRTYPE, LNKTYPE, SYMTYPE, TarError, TarInfo
from tarfile import open as tar_open
from typing import Protocol

import pytest

from build._compat.tarfile import _validate_safe_member, safe_extractall


FileMember = Callable[[str, bytes], TarInfo]
DeviceMember = Callable[[str], TarInfo]
ArchiveBuilder = Callable[[Path, 'list[tuple[TarInfo, bytes | None]]'], None]


class LinkMember(Protocol):
    def __call__(self, name: str, linkname: str, *, hard: bool = ...) -> TarInfo: ...


def test_safe_extractall_extracts_clean_archive(
    tmp_path: Path,
    make_archive: ArchiveBuilder,
    file_member: FileMember,
) -> None:
    archive = tmp_path / 'clean.tar'
    body = b'meta'
    make_archive(archive, [(file_member('pkg/PKG-INFO', body), body)])

    out = tmp_path / 'out'
    out.mkdir()
    with tar_open(archive) as tar:
        safe_extractall(tar, str(out))

    assert (out / 'pkg' / 'PKG-INFO').read_bytes() == body


def test_validate_safe_member_accepts_clean_file(tmp_path: Path, file_member: FileMember) -> None:
    base = tmp_path.resolve()
    _validate_safe_member(file_member('pkg/file.txt', b'x'), base)


def test_validate_safe_member_rejects_path_traversal(tmp_path: Path, file_member: FileMember) -> None:
    base = tmp_path.resolve()
    with pytest.raises(TarError, match='escapes destination'):
        _validate_safe_member(file_member('../evil.txt', b'x'), base)


def test_validate_safe_member_rejects_absolute_symlink(tmp_path: Path, link_member: LinkMember) -> None:
    base = tmp_path.resolve()
    with pytest.raises(TarError, match='link target escapes'):
        _validate_safe_member(link_member('pkg/evil', '/etc/passwd'), base)


def test_validate_safe_member_rejects_relative_escape_symlink(tmp_path: Path, link_member: LinkMember) -> None:
    base = tmp_path.resolve()
    with pytest.raises(TarError, match='link target escapes'):
        _validate_safe_member(link_member('pkg/evil', '../../outside'), base)


def test_validate_safe_member_accepts_safe_symlink(tmp_path: Path, link_member: LinkMember) -> None:
    base = tmp_path.resolve()
    _validate_safe_member(link_member('pkg/link', 'real.txt'), base)


def test_validate_safe_member_rejects_escaping_hardlink(tmp_path: Path, link_member: LinkMember) -> None:
    base = tmp_path.resolve()
    with pytest.raises(TarError, match='link target escapes'):
        _validate_safe_member(link_member('pkg/evil', '../../outside', hard=True), base)


def test_validate_safe_member_rejects_device_file(tmp_path: Path, device_member: DeviceMember) -> None:
    base = tmp_path.resolve()
    with pytest.raises(TarError, match='special device file'):
        _validate_safe_member(device_member('pkg/null'), base)


@pytest.fixture
def make_archive() -> ArchiveBuilder:
    def _build(path: Path, members: list[tuple[TarInfo, bytes | None]]) -> None:
        with tar_open(path, 'w') as tar:
            for info, body in members:
                tar.addfile(info, BytesIO(body) if body is not None else None)

    return _build


@pytest.fixture
def file_member() -> FileMember:
    def _build(name: str, body: bytes) -> TarInfo:
        info = TarInfo(name=name)
        info.size = len(body)
        return info

    return _build


@pytest.fixture
def link_member() -> LinkMember:
    def _build(name: str, linkname: str, *, hard: bool = False) -> TarInfo:
        info = TarInfo(name=name)
        info.type = LNKTYPE if hard else SYMTYPE
        info.linkname = linkname
        return info

    return _build


@pytest.fixture
def device_member() -> DeviceMember:
    def _build(name: str) -> TarInfo:
        info = TarInfo(name=name)
        info.type = CHRTYPE
        info.devmajor = 1
        info.devminor = 3
        return info

    return _build
